import unittest
from collections.abc import Iterable
from unittest.mock import MagicMock

from dihub.__internal.proxies import ProviderProxy
from dihub.decorators import provider
from dihub.exceptions import NotAPyDIProvider, ProviderNotFound
from .provider_delegate import ProviderDelegate


class ProviderDelegateTest(unittest.TestCase):
    def test___init___should_raise_NotAPyDIProvider(self):
        def p(): ...

        providers = [p]
        with self.assertRaises(NotAPyDIProvider):
            ProviderDelegate(providers)

    def test___init___should_append_providers(self):
        @provider
        def p(): ...

        providers = [p]
        expected = ProviderDelegate(providers)
        self.assertEqual(len(expected), 1)

    def test___add___should_concat_providers(self):
        @provider
        def p(): ...

        providers = [p]
        expected = ProviderDelegate(providers)

        others = ProviderDelegate(providers)

        initial = len(expected)

        expected += providers
        self.assertEqual(len(expected), initial + len(providers))

        initial = len(expected)

        expected += others
        self.assertEqual(len(expected), initial + len(others))

    def test_append_should_append_providers(self):
        @provider
        def p(): ...

        providers = [p]
        expected = ProviderDelegate(providers)
        initial = len(expected)

        expected.append(p)
        self.assertEqual(len(expected), initial + 1)

    def test_ProviderDelegate_should_iterable(self):
        self.assertTrue(issubclass(ProviderDelegate, Iterable))

    @staticmethod
    def test_on_boot_should_call_provider_on_boot():
        @provider
        def p(): ...

        providers = [p]
        expected = ProviderDelegate(providers)
        expected[p.__name__][0].on_boot = MagicMock()

        expected.on_boot()
        expected[p.__name__][0].on_boot.assert_called_once()

    @staticmethod
    def test_on_post_boot_should_call_provider_on_post_boot():
        @provider
        def p(): ...

        providers = [p]
        expected = ProviderDelegate(providers)
        expected[p.__name__][0].on_post_boot = MagicMock()

        expected.on_post_boot()
        expected[p.__name__][0].on_post_boot.assert_called_once()

    def test___getitem___should_raise_ProviderNotFound_when_no_token_match(self):
        expected = ProviderDelegate([])
        with self.assertRaises(ProviderNotFound):
            expected['any']

    def test___getitem___should_return_match_provider(self):
        @provider
        def p(): ...

        pd = ProviderDelegate([p])
        expected, _ = pd[p.__name__]

        self.assertIsInstance(expected, ProviderProxy)
        self.assertIs(p, expected.release())

    def test___str___should_return_str(self):
        @provider
        def p(): ...

        pd = ProviderDelegate([p])

        self.assertIsInstance(str(pd), str)

    def test___delitem___should_remove_provider(self):
        @provider
        def p(): ...

        pd = ProviderDelegate([p])

        initial = len(pd)
        del pd["p"]
        self.assertEqual(len(pd), initial - 1)


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from inspect import getmembers, isclass
from typing import Any, Type

from dihub.__internal.helpers import get_class_name
from dihub.types import Value, IProviderProxy, IProviderRunner


class ProviderProxy(IProviderProxy):
    __provide: Value
    __AFTER_START_CALLED = False

    def __init__(self, provide: Value):
        self.__provide = provide

    def __str__(self):
        return "<%s %s>" % (ProviderProxy.__name__, get_class_name(self.__provide))

    def __eq__(self, other: Any):
        if isinstance(other, ProviderProxy):
            return self.__provide == other.__provide

        return False

    def deep_eq(self, other: Any) -> bool:
        if not self.__eq__(other):
            return False

        return id(self.__provide) == id(other.__provide)

    def on_boot(self):
        if isclass(self.__provide):
            for _, method in getmembers(self.__provide, predicate=lambda x: isinstance(x, ProviderProxy)):
                method.on_boot()
            self.__provide = self.__provide()

    def on_post_boot(self):
        if not isclass(self.__provide):
            for _, method in getmembers(self.__provide, predicate=lambda x: isinstance(x, ProviderProxy)):
                method.on_post_boot()

            if not self.__AFTER_START_CALLED and isinstance(self.__provide, IProviderRunner):
                self.__provide.after_started()
                self.__AFTER_START_CALLED = True

    def cast(self, metaclass: Type[Value]) -> Value:
        if isinstance(self.__provide, metaclass):
            return self
        raise TypeError("Can't cast, the provider %s doesn't implement '%s'" % (self.__provide, metaclass.__name__))

    def release(self) -> Value:
        if isinstance(self.__provide, ProviderProxy):
            return self.__provide.release()

        return self.__provide

    def __getattr__(self, item):
        return getattr(self.__provide, item)

    def __copy__(self):
        # Avoid copy ProviderProxy.provide as ProviderProxy
        if isinstance(self.__provide, ProviderProxy):
            return self.__provide.__copy__()

        return ProviderProxy(deepcopy(self.__provide))

    @property
    def provide(self) -> Value:
        return self.__provide



from dihub.__internal.helpers import get_inject_token_str
from dihub.types import InjectToken


class InjectedDelegate:
    __token: InjectToken

    def __init__(self, token: InjectToken):
        self.__token = token

    @property
    def token(self) -> str:
        return get_inject_token_str(self.__token)



from inspect import isfunction
from typing import Optional

from dihub.__internal.helpers import AnnotationOf, get_class_name
from dihub.__internal.proxies import PrimitiveProxy
from dihub.constants import ProviderScope, _PROVIDER_ANNOTATIONS, ROOT_MODULE_DELEGATE
from dihub.exceptions import ReservedInjectToken
from dihub.types import Value, ProviderAnnotation


def __process_provider_decorator(provide: Value, token: str, scope: ProviderScope):
    if isfunction(provide):
        provide = PrimitiveProxy(provide.__name__, provide())

    final_token = token
    if final_token is None:
        final_token = get_class_name(provide)

    if final_token == ROOT_MODULE_DELEGATE:
        raise ReservedInjectToken(final_token)

    AnnotationOf(provide).set(_PROVIDER_ANNOTATIONS, ProviderAnnotation(token=final_token, scope=scope))
    return provide


def provider(provide: Optional[Value] = None, /, *, token: Optional[str] = None, scope: ProviderScope = ProviderScope.GLOBAL):
    """
    :param provide: The decorated class
    :param token: Provider injects token. Used to refer the provider upon injection
    :param scope: The provider scopes
    """

    def wrapper(cls: Value) -> Value:
        return __process_provider_decorator(cls, token, scope)

    if provide is None:
        return wrapper

    return wrapper(provide)



from inspect import getmembers
from typing import Optional, Self, Any, Tuple, List

from dihub.__internal.helpers import AnnotationOf, validate_dihub_module, get_class_name
from dihub.__internal.proxies import ProviderProxy
from dihub.constants import _MODULE_ANNOTATIONS, ProviderScope, ROOT_MODULE_DELEGATE
from dihub.exceptions import (
    CannotResolveDependency,
    ModuleNotFound,
    ProviderNotFound,
)
from dihub.types import (
    IModuleDelegate,
    IProviderDelegate,
    ModuleAnnotation,
    ProviderAnnotation,
    Instance,
)

from .injected_delegate import InjectedDelegate
from .provider_delegate import ProviderDelegate


class ModuleDelegate(IModuleDelegate):
    __base_class: type
    __providers: ProviderDelegate
    __imported_modules_delegate: List[IModuleDelegate]
    __parent_delegate: Self
    __for_root_imports: List[IModuleDelegate]

    def __init__(self, module: type, parent_delegate: Optional[Self]):
        validate_dihub_module(module)

        annotations = AnnotationOf(module).get(
            _MODULE_ANNOTATIONS, ModuleAnnotation
        )

        self.__base_class = module
        self.__parent_delegate = parent_delegate
        self.__providers = ProviderDelegate(annotations.providers)
        self.__imported_modules_delegate = [ModuleDelegate(m, self) for m in annotations.imports]

        self.__for_root_imports = []
        for md in self.__imported_modules_delegate:
            md_annotations = AnnotationOf(md.base_class).get(_MODULE_ANNOTATIONS, ModuleAnnotation)
            if md_annotations.for_root:
                self.__for_root_imports.append(md)

    @property
    def is_root(self) -> bool:
        return self.__parent_delegate is None

    def __call__(self, *args, **kwargs):
        return self

    def __str__(self):
        imports_str = []
        for i in self.__imported_modules_delegate:
            imports_str.append(get_class_name(i))
        return "<%s %s<imports [%s]>>" % (
            ModuleDelegate.__name__,
            get_class_name(self.__base_class),
            ",\n".join(imports_str),
        )

    def __eq__(self, other: Any):
        if isinstance(other, ModuleDelegate):
            return self.__base_class == other.__base_class
        elif isinstance(other, type):
            return self.__base_class == other

        return False

    def __setattr__(self, key, value):
        super(ModuleDelegate, self).__setattr__(key, value)
        setattr(self.__base_class, key, value)

    def deep_eq(self, other: Any) -> bool:
        if not self.__eq__(other):
            return False
        return id(other) == id(self)

    def __getitem__(self, module: Instance) -> IModuleDelegate:
        for m in self.__imported_modules_delegate:
            if m == module:
                return m
        raise ModuleNotFound(module)

    def get_exported_provider(
            self, token: str
    ) -> Tuple[Optional[ProviderProxy], Optional[ProviderAnnotation]]:
        provider, annotations = self.__providers[token]
        if provider is not None and annotations.exported:
            if annotations.scope == ProviderScope.MODULE:
                return provider.__copy__(), annotations

            return provider, annotations

        raise ProviderNotFound(token)

    def get_for_root_provider(self, token: str) -> Tuple[Optional[ProviderProxy], Optional[ProviderAnnotation]]:
        for rmd in self.__for_root_imports:
            try:
                dependency, annotations = rmd.get_exported_provider(token)
                return dependency, annotations
            except ProviderNotFound:
                continue

        if self.parent_delegate is None:
            """
                it's already on root
                Search from root providers. It's special ability for root module that
                all of the root exported providers can be accessed by the importd module providers 
            """
            return self.get_exported_provider(token)

        return self.parent_delegate.get_for_root_provider(token)

    def __resolve_providers_dependencies(self):
        def resolve_root_module_delegate(provider: ProviderProxy, attribute_name: str, injected_delegate: InjectedDelegate) -> bool:
            if injected_delegate.token == ROOT_MODULE_DELEGATE:
                setattr(provider.provide, attribute_name, self.root_delegate)

                return True

            return False

        for provider, _ in self.__providers:
            for dependency_name, injected_dependency in getmembers(
                    provider.provide, predicate=lambda x: isinstance(x, InjectedDelegate)
            ):
                if resolve_root_module_delegate(provider, dependency_name, injected_dependency):
                    continue

                dependency: Optional[ProviderProxy] = None
                annotations: Optional[ProviderAnnotation] = None

                # Resolve from self
                try:
                    dependency, annotations = self.__providers[
                        injected_dependency.token
                    ]
                except ProviderNotFound:
                    pass

                if dependency is None:
                    for m in self.__imported_modules_delegate:
                        try:
                            dependency, annotations = m.get_exported_provider(
                                injected_dependency.token
                            )
                            break
                        except ProviderNotFound:
                            continue

                if not self.is_root and dependency is None:
                    try:
                        dependency, annotations = self.parent_delegate.get_for_root_provider(injected_dependency.token)
                    except ProviderNotFound:
                        raise CannotResolveDependency(dependency_name, str(provider))
                else:
                    raise CannotResolveDependency(dependency_name, str(provider))

                # Decide how the dependency is resolved
                if annotations.scope == ProviderScope.LOCAL:
                    # Make the dependency uniqueness by copy
                    dependency = dependency.__copy__()
                else:
                    self.__providers.append(dependency)

                setattr(provider.provide, dependency_name, dependency)

    def __clean_up(self):
        # Delete all LOCAL scoped providers
        for p, annotations in self.__providers:
            if annotations.scope == ProviderScope.LOCAL:
                del self.__providers[annotations.token]

    def on_boot(self):
        # Recursive boot imported module
        for i in self.__imported_modules_delegate:
            i.on_boot()

        self.__providers.on_boot()

        # Injecting
        self.__resolve_providers_dependencies()

    def on_post_boot(self):
        for i in self.__imported_modules_delegate:
            i.on_post_boot()

        self.__providers.on_post_boot()

        self.__clean_up()

    @property
    def providers(self) -> IProviderDelegate:
        return self.__providers

    @property
    def base_class(self) -> type:
        return self.__base_class

    @property
    def root_delegate(self) -> Self:
        if not self.is_root:
            return self.parent_delegate.root_delegate

        return self

    @property
    def parent_delegate(self) -> Self:
        return self.__parent_delegate

    @property
    def imports(self) -> List[Self]:
        return self.__imported_modules_delegate



from typing import Any

from dihub.constants import _PROVIDER_ANNOTATIONS, _MODULE_ANNOTATIONS
from dihub.exceptions import NotAPyDIProvider, NotAPyDIModule
from dihub.types import ProviderAnnotation, ModuleAnnotation
from .annotation_of import AnnotationOf


def __generic_validate_dihub(t: Any, ann_name: str, expected_type: Any, not_match_exception: Exception):
    try:
        if AnnotationOf(t).get(ann_name, expected_type) is None:
            raise not_match_exception
    except ValueError:
        raise not_match_exception


def validate_dihub_provider(t: Any):
    __generic_validate_dihub(t, _PROVIDER_ANNOTATIONS, ProviderAnnotation, NotAPyDIProvider(str(t)))


def validate_dihub_module(t: Any):
    __generic_validate_dihub(t, _MODULE_ANNOTATIONS, ModuleAnnotation, NotAPyDIModule(str(t)))



from abc import abstractmethod, ABC
from dataclasses import dataclass
from types import FunctionType, LambdaType
from typing import TypeAlias, TypeVar, Callable, List, Union, Self, Tuple, Optional, Any, Type, Protocol

from .constants import ProviderScope

Value = TypeVar("Value", covariant=True)
Config = TypeVar("Config", contravariant=True)


class Configurable(Protocol[Config, Value]):
    @classmethod
    @abstractmethod
    def configure(cls, Config: Config) -> Value: ...

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Value: ...


Instance = TypeVar('Instance', type, Configurable, covariant=True)
Method: TypeAlias = Callable[[Instance], Value]

NOTABLE = Union[type, FunctionType, LambdaType]

InjectToken = Union[str, type]

_as_class: TypeAlias = type
_as_primitive: TypeAlias = Union[str, int, bool, float]

Provide = TypeVar("Provide", covariant=True)

Modules: TypeAlias = List[Instance]


def is_notable(obj) -> bool:
    return isinstance(obj, FunctionType) or isinstance(obj, LambdaType) or isinstance(obj, type)


class IProviderProxy(ABC):
    @abstractmethod
    def on_boot(self): ...

    @abstractmethod
    def on_post_boot(self): ...

    @abstractmethod
    def deep_eq(self, other: Any) -> bool: ...

    @abstractmethod
    def cast(self, metaclass: Type[Value]) -> Value: ...

    @abstractmethod
    def release(self) -> Value: ...


Providers: TypeAlias = List[Provide]


@dataclass(frozen=True)
class ProviderAnnotation:
    token: str
    scope: ProviderScope
    exported: bool = False


class IProviderDelegate(ABC):
    @abstractmethod
    def __getitem__(self, token: InjectToken) -> Tuple[IProviderProxy, ProviderAnnotation]: ...


class IModuleDelegate(ABC):
    @property
    @abstractmethod
    def is_root(self) -> bool: ...

    @abstractmethod
    def __getitem__(self, module: Instance) -> Self: ...

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Self: ...

    @abstractmethod
    def deep_eq(self, other: Any) -> bool: ...

    @property
    @abstractmethod
    def providers(self) -> IProviderDelegate: ...

    @property
    @abstractmethod
    def imports(self) -> List[Self]: ...

    @property
    @abstractmethod
    def base_class(self) -> type: ...

    @property
    @abstractmethod
    def root_delegate(self) -> Self: ...

    @property
    @abstractmethod
    def parent_delegate(self) -> Self: ...

    @abstractmethod
    def on_boot(self): ...

    @abstractmethod
    def on_post_boot(self): ...

    @abstractmethod
    def get_exported_provider(self, token: str) -> Tuple[Optional[IProviderProxy], Optional[ProviderAnnotation]]: ...

    @abstractmethod
    def get_for_root_provider(self, token: str) -> Tuple[Optional[IProviderProxy], Optional[ProviderAnnotation]]: ...


@dataclass(frozen=True)
class ModuleAnnotation:
    imports: Modules
    providers: Providers
    for_root: bool = False


class IRootRunner(ABC):
    def __await__(self): ...

    @abstractmethod
    def after_started(self, root_module_delegate: IModuleDelegate): ...


class IConfigurableModule(Protocol[Value]):
    __config: Value

    @property
    def Config(self) -> Value:
        return self.__config

    def configure(self, config: Value):
        self.__config = config


class IRootPlugin(ABC):
    @abstractmethod
    def __call__(self, root_module_delegate: IModuleDelegate): ...


Plugins = List[IRootPlugin]


class IProviderRunner(ABC):
    @abstractmethod
    def after_started(self): ...



from copy import deepcopy
from typing import Callable, Type

from dihub.__internal.helpers import AnnotationOf, get_class_name
from dihub.constants import _MODULE_ANNOTATIONS
from dihub.types import Value, ModuleAnnotation, Configurable, Config
from .provider import provider


def configurable(config_type: Type[Config]) -> Callable[[Value], Configurable[Config, Value]]:
    def wrapper(decorated_class: Value) -> Configurable[Config, Value]:
        class Wrapper(decorated_class):
            def __init__(self, *args, **kwargs):
                super(decorated_class, self).__init__(*args, **kwargs)

            def __eq__(self, other):
                return super().__eq__(other)

            @classmethod
            def configure(cls, config: Config) -> Value:
                if not isinstance(config, config_type):
                    raise TypeError("Config not an instance of '%s'" % get_class_name(config_type))

                @provider(token=get_class_name(config_type))
                def config_wrapper() -> Config:
                    return config

                class Caller:
                    def __call__(self, *args, **kwargs):
                        return decorated_class(*args, **kwargs)

                Caller.__annotations__ = deepcopy(cls.__annotations__)

                module_annotation = AnnotationOf(Caller).get(_MODULE_ANNOTATIONS, ModuleAnnotation)
                module_annotation.providers.append(config_wrapper)

                return Caller()

        Wrapper.__annotations__.update(deepcopy(decorated_class.__annotations__))
        return Wrapper

    return wrapper



class NotAClassException(ValueError):
    def __init__(self, name: str):
        super().__init__("this '%s' is not a Class" % name)


class MissingArgumentException(ValueError):
    def __init__(self, arg_name: str):
        super().__init__("missing '%s'" % arg_name)


class NotAPyDIModule(ValueError):
    def __init__(self, name: str):
        super().__init__("this: %s, not a dihub module" % name)


class NotAPyDIProvider(ValueError):
    def __init__(self, name: str):
        super().__init__("this: %s, not a dihub provider" % name)


class AnnotationsNotSupported(TypeError):
    def __init__(self, name: str):
        super().__init__("this: %s, not support annotations" % name)


class CannotResolveDependency(RuntimeError):
    def __init__(self, dep_name: str, at: str):
        super().__init__("Cannot resolve '%s' at '%s'" % (dep_name, at))


class ProviderNotFound(KeyError):
    def __init__(self, token: str):
        super().__init__("Provider with token '%s' not found" % token)


class ModuleNotFound(KeyError):
    def __init__(self, module: type):
        super().__init__("Module '%s' not found" % str(module))


class ReservedInjectToken(ValueError):
    def __init__(self, token: str):
        super().__init__("Inject token '%s' is reserved" % token)



from typing import Union, Tuple, List

from dihub.__internal.helpers import AnnotationOf, validate_dihub_provider, get_class_name, get_inject_token_str
from dihub.__internal.proxies import ProviderProxy
from dihub.constants import _PROVIDER_ANNOTATIONS
from dihub.exceptions import ProviderNotFound
from dihub.types import Providers, InjectToken, IProviderDelegate, ProviderAnnotation, Provide


class ProviderDelegate(IProviderDelegate):
    __providers: List[ProviderProxy]
    __iter_index = -1

    def __init__(self, providers: Providers):
        self.__providers = []
        for i in providers:
            validate_dihub_provider(i)

            self.__providers.append(ProviderProxy(i))

    def __str__(self):
        members_str = []
        for i in self.__providers:
            annotations = AnnotationOf(i.provide).get(_PROVIDER_ANNOTATIONS, ProviderAnnotation)
            members_str.append(
                "'%s': '%s'" % (annotations.token, get_class_name(i.provide)))
        return "{%s}" % (", ".join(members_str))

    def __add__(self, other: Union[IProviderDelegate, Providers]):
        if isinstance(other, ProviderDelegate):
            self.__providers += other.__providers
        else:
            for p in other:
                validate_dihub_provider(p)
                self.__providers.append(ProviderProxy(p))

        return self

    def __len__(self):
        return len(self.__providers)

    def append(self, item: Union[ProviderProxy, Provide]):
        if isinstance(item, ProviderProxy):
            self.__providers.append(item)
        else:
            validate_dihub_provider(item)
            self.__providers.append(ProviderProxy(item))
        return self

    def __iter__(self):
        self.__iter_index = -1
        return self

    def __next__(self) -> Tuple[ProviderProxy, ProviderAnnotation]:  # Python 2: def next(self)
        if self.__iter_index < len(self.__providers) - 1:
            self.__iter_index += 1
            p = self.__providers[self.__iter_index]
            annotations = AnnotationOf(p.provide).get(_PROVIDER_ANNOTATIONS, ProviderAnnotation)
            return p, annotations
        raise StopIteration

    def __is_token_match(self, this_token: str, that_token: InjectToken) -> bool:
        return get_inject_token_str(that_token) == this_token

    def __getitem__(self, token: InjectToken) -> Tuple[ProviderProxy, ProviderAnnotation]:
        for p in self.__providers:
            annotations = AnnotationOf(p.provide).get(_PROVIDER_ANNOTATIONS, ProviderAnnotation)

            if self.__is_token_match(annotations.token, token):
                return p, annotations

        raise ProviderNotFound(token)

    def __delitem__(self, token: InjectToken):
        for i, p in enumerate(self.__providers):
            annotations = AnnotationOf(p.provide).get(_PROVIDER_ANNOTATIONS, ProviderAnnotation)
            if self.__is_token_match(annotations.token, token):
                del self.__providers[i]
                return

    def on_boot(self):
        # Construct
        for i in self.__providers:
            i.on_boot()

    def on_post_boot(self):
        for i in self.__providers:
            i.on_post_boot()



from typing import Any, Type, Optional

from dihub.exceptions import AnnotationsNotSupported
from dihub.types import Value, NOTABLE
from .class_helper import get_class_name


def is_primitive(obj):
    return not hasattr(obj, '__dict__')


class AnnotationOf:
    __source: NOTABLE

    def __init__(self, source: NOTABLE):
        self.__source = source

    def get(self, ann_name: str, metaclass: Optional[Type[Value]] = None) -> Optional[Value]:
        try:
            value = self.__source.__annotations__[ann_name]
            if metaclass is None:
                return value

            if not isinstance(value, metaclass):
                raise ValueError("The annotation values doesn't implement '%s'. Actual '%s'" % (
                    get_class_name(metaclass), get_class_name(value, True)))

            return value
        except AttributeError:
            raise AnnotationsNotSupported(str(self.__source))
        except KeyError:
            return None

    def set(self, ann_name: str, value: Any):
        try:
            self.__source.__annotations__[ann_name] = value
        except AttributeError:
            raise AnnotationsNotSupported(str(self.__source))



from dihub.__internal.helpers import AnnotationOf, validate_dihub_module

from dihub.constants import _MODULE_ANNOTATIONS
from dihub.types import ModuleAnnotation, Instance


def for_root(m: Instance) -> Instance:
    validate_dihub_module(m)

    module_annotations = AnnotationOf(m).get(_MODULE_ANNOTATIONS, ModuleAnnotation)
    AnnotationOf(m).set(_MODULE_ANNOTATIONS, ModuleAnnotation(imports=module_annotations.imports, providers=module_annotations.providers,
                                                              for_root=True))

    return m



from dihub.__internal.helpers import AnnotationOf, validate_dihub_provider

from dihub.constants import _PROVIDER_ANNOTATIONS
from dihub.types import ProviderAnnotation, Value


def export(provide: Value) -> Value:
    validate_dihub_provider(provide)

    provider_annotations = AnnotationOf(provide).get(_PROVIDER_ANNOTATIONS, ProviderAnnotation)
    AnnotationOf(provide).set(_PROVIDER_ANNOTATIONS, ProviderAnnotation(token=provider_annotations.token, scope=provider_annotations.scope,
                                                                        exported=True))
    return provide



from typing import Any, Generic

from dihub.types import _as_primitive, Value


class PrimitiveProxy(Generic[Value]):
    __name: str
    __primitive_value: _as_primitive

    def __init__(self, name: str, primitive_value: _as_primitive):
        self.__primitive_value = primitive_value
        self.__name = name

    @property
    def __name__(self):
        return self.__name

    def __get__(self, *args, **kwargs) -> _as_primitive:
        return self.__primitive_value

    def __str__(self):
        return self.__primitive_value.__str__()

    def __eq__(self, other: Any):
        return self.__primitive_value.__eq__(other)

    def __repr__(self):
        return self.__primitive_value.__repr__()

    def set_actual_value(self, value: Value):
        self.__primitive_value = value



from .configurable import configurable
from .export import export
from .for_root import for_root
from .inject import inject
from .module import module
from .provider import provider
from .root import root



from typing import Any

from dihub.types import InjectToken


def get_class_name(obj: Any, class_first=False) -> str:
    if class_first:
        if hasattr(obj, '__class__'):
            return obj.__class__.__name__
        elif hasattr(obj, '__name__'):
            return obj.__name__
    else:
        if hasattr(obj, '__name__'):
            return obj.__name__
        elif hasattr(obj, '__class__'):
            return obj.__class__.__name__

    return type(obj).__name__


def get_inject_token_str(token: InjectToken) -> str:
    if isinstance(token, str):
        return token
    return get_class_name(token)



from enum import Enum

_MODULE_ANNOTATIONS = "__PYDI_MODULE"
_PROVIDER_ANNOTATIONS = "__PYDI_PROVIDER"
ROOT_MODULE_DELEGATE = "__PYDI_ROOT_MODULE_DELEGATE"


class ProviderScope(Enum):
    GLOBAL = 1, 'Provider is only one SINGLE instance (singleton) which shared to all module imports'
    MODULE = 2, 'Provider is UNIQUE instance on each module imports'
    LOCAL = 3, 'Provider is UNIQUE instance on each providers'



import unittest
from unittest.mock import MagicMock, AsyncMock

from dihub.__internal.delegates import InjectedDelegate
from dihub.constants import _MODULE_ANNOTATIONS, _PROVIDER_ANNOTATIONS, ROOT_MODULE_DELEGATE
from dihub.decorators import module, inject, provider, root, export, for_root
from dihub.exceptions import NotAPyDIModule, NotAPyDIProvider, ReservedInjectToken
from dihub.types import ModuleAnnotation, ProviderAnnotation, IRootPlugin, IRootRunner, IModuleDelegate


class DecoratorsTest(unittest.IsolatedAsyncioTestCase):
    def test_module_decorator_should_set_module_annotation(self):
        @module
        class C: ...

        self.assertIn(_MODULE_ANNOTATIONS, C.__annotations__)
        self.assertIsInstance(C.__annotations__[_MODULE_ANNOTATIONS], ModuleAnnotation)

        default_annotation = ModuleAnnotation(imports=[], providers=[])
        self.assertEqual(default_annotation, C.__annotations__[_MODULE_ANNOTATIONS])

    def test_inject_should_return_InjectedDelegate(self):
        class C:
            member: str = inject("token")

        c = C()
        self.assertIsInstance(c.member, InjectedDelegate)
        self.assertEqual(c.member.token, "token")

    def test_provider_decorator_should_set_provider_annotation(self):
        @provider(token="token")
        class C: ...

        self.assertIn(_PROVIDER_ANNOTATIONS, C.__annotations__)
        self.assertIsInstance(C.__annotations__[_PROVIDER_ANNOTATIONS], ProviderAnnotation)

        @provider(token="provide")
        def constant_provider() -> str: ...

        self.assertIn(_PROVIDER_ANNOTATIONS, constant_provider.__annotations__)
        self.assertIsInstance(constant_provider.__annotations__[_PROVIDER_ANNOTATIONS], ProviderAnnotation)

        @provider
        class B: ...

        self.assertIn(_PROVIDER_ANNOTATIONS, B.__annotations__)
        self.assertIsInstance(B.__annotations__[_PROVIDER_ANNOTATIONS], ProviderAnnotation)
        self.assertIs(B.__annotations__[_PROVIDER_ANNOTATIONS].token, B.__name__)

        b = B()
        provider(b)
        self.assertIn(_PROVIDER_ANNOTATIONS, b.__annotations__)
        self.assertIsInstance(b.__annotations__[_PROVIDER_ANNOTATIONS], ProviderAnnotation)
        self.assertIs(b.__annotations__[_PROVIDER_ANNOTATIONS].token, b.__class__.__name__)

    def test_provider_decorator_should_raise_ReservedInjectToken(self):
        with self.assertRaises(ReservedInjectToken):
            @provider(token=ROOT_MODULE_DELEGATE)
            class B: ...

    def test_root_decorator_should_raise_NotAPyDIModule_to_non_module(self):
        with self.assertRaises(NotAPyDIModule):
            @root
            class C: ...

    def test_root_decorator_should_not_affect_class_initialization(self):
        class OC: ...

        # Implicitly decorate
        C = module()(OC)
        C = root(C)

        c1 = C()
        self.assertIsInstance(c1, OC)

        c2 = C()
        self.assertIsInstance(c2, OC)

        self.assertNotEqual(c1, c2)

    def test_root_decorator_should_wrap_a_class(self):
        class OC: ...

        # Implicitly decorate
        C = module()(OC)
        C = root(C)

        c = C()
        self.assertIsInstance(c, OC)

    def test_root_decorator_should_call_plugin(self):
        class Plugin(IRootPlugin):
            def __call__(self, *args, **kwargs):
                pass

        Plugin.__call__ = MagicMock()

        class OC: ...

        # Implicitly decorate
        C = module()(OC)
        C = root(C, plugins=[Plugin()])

        c = C()
        Plugin.__call__.assert_called_once()

    async def test_root_decorator_should_after_started(self):
        class OC(IRootRunner):
            async def after_started(self, root_module_delegate: IModuleDelegate):
                pass

        OC.after_started = AsyncMock()

        # Implicitly decorate
        C = module()(OC)
        C = root()(C)

        c = C()
        await c
        OC.after_started.assert_called_once()

    def test_root_decorator_should_not_affect_the_actual_class(self):
        class OC:
            def __init__(self, an_arg: int): ...

            def a_method(self): ...

        OC.a_method = MagicMock()
        OC.__init__ = MagicMock(return_value=None)

        # Implicitly decorate
        C = module()(OC)
        C = root(C)

        c = C(an_arg=1)
        OC.__init__.assert_called_once_with(an_arg=1)

        c.a_method()
        OC.a_method.assert_called_once()

    def test_export_decorator_should_set_annotation_to_exported(self):
        @export
        @provider(token="token")
        class C: ...

        annotations: ProviderAnnotation = C.__annotations__.get(_PROVIDER_ANNOTATIONS)
        self.assertIs(annotations.exported, True)

    def test_export_decorator_should_raise_NotAPyDIProvider_to_non_provider(self):
        with self.assertRaises(NotAPyDIProvider):
            @export
            class C: ...

    def test_for_root_decorator_should_set_annotation_to_for_root(self):
        @for_root
        @module
        class C: ...

        annotations: ModuleAnnotation = C.__annotations__.get(_MODULE_ANNOTATIONS)
        self.assertIs(annotations.for_root, True)

    def test_export_decorator_should_raise_NotAPyDIModule_to_non_module(self):
        with self.assertRaises(NotAPyDIModule):
            @for_root
            class C: ...


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Awaitable, Optional

from dihub.types import IRootPlugin, IModuleDelegate, InjectToken


class ASGI(IRootPlugin):
    from_provider: InjectToken
    from_module: Optional[type] = None
    asgi_provider: Callable[..., Awaitable[None]]

    def __init__(self, from_provider: InjectToken, from_module: Optional[type] = None):
        self.from_module = from_module
        self.from_provider = from_provider

    async def __asgi_async_call__(self, *args, **kwargs):
        return await self.asgi_provider(*args, **kwargs)

    def __call__(self, root_module_delegate: IModuleDelegate):
        if not root_module_delegate.is_root:
            raise ValueError("module is not root")

        if self.from_module is not None:
            self.asgi_provider = root_module_delegate[self.from_module].providers[self.from_provider][0].release()
        else:
            self.asgi_provider = root_module_delegate.providers[self.from_provider][0].release()

        setattr(root_module_delegate, "__call__", self.__asgi_async_call__)



from inspect import iscoroutinefunction
from typing import Type

from dihub.__internal.delegates import ModuleDelegate
from dihub.__internal.helpers import validate_dihub_module
from dihub.types import Value, IRootRunner, Plugins


def __process_root_decorator(cls: Value, plugins: Plugins) -> Type[Value]:
    validate_dihub_module(cls)

    def __call__(*args, **kwargs):
        __module_delegate = ModuleDelegate(cls, None)
        __module_delegate.on_boot()
        __module_delegate.on_post_boot()

        # Apply plugins
        for p in plugins:
            p(__module_delegate)

        base_class_instance = __module_delegate.base_class(*args, **kwargs)

        if isinstance(base_class_instance, IRootRunner):
            if iscoroutinefunction(base_class_instance.after_started):
                async def after_started_coro(instance: IRootRunner):
                    await instance.after_started(__module_delegate)
                    return instance

                return after_started_coro(base_class_instance)
            else:
                base_class_instance.after_started(__module_delegate)

        return base_class_instance

    return __call__


def root(cls: Value = None, /, *, plugins: Plugins = None) -> Type[Value]:
    if plugins is None:
        plugins = []

    def wrapper(_cls) -> Type[Value]:
        return __process_root_decorator(_cls, plugins)

    if cls is None:
        return wrapper

    return wrapper(cls)



from typing import List, Any, Callable, Union, Optional

from dihub.__internal.helpers import AnnotationOf
from dihub.constants import _MODULE_ANNOTATIONS
from dihub.types import Value, Providers, Modules, ModuleAnnotation


def __get_list_defaults(arr: List[Any]) -> List[Any]:
    if arr is None:
        return []

    return arr


def __process_module_decorator(cls: Value, imports: Modules, providers: Providers) -> Value:
    AnnotationOf(cls).set(_MODULE_ANNOTATIONS, ModuleAnnotation(
        imports=imports,
        providers=providers))

    return cls


def module(cls: Optional[Value] = None, /, *, imports: Optional[Modules] = None, providers: Optional[Providers] = None) -> Union[
    Callable[[Value], Value], Value]:
    def wrapper(_cls: Value) -> Value:
        return __process_module_decorator(_cls, __get_list_defaults(imports), __get_list_defaults(providers))

    if cls is None:
        return wrapper

    return wrapper(cls)
